/* bigcal.c - make big calendars
 *
 *	Copyright (c) 1990 Tony Aiuto
 *
 * Edit History:
 * 07/23/91 tony - created
 * 12/05/91 tony - better formatting, add edit option
 * 12/10/91 tony - bug fix, wrap months back to 1 rather than 13
 * 02/26/93 tony - fix next month lookup for spanning three months.
 * 03/26/93 tony - print ~/.todo at end
 * 10/28/93 tony - read multiple calandars
 * 10/28/93 tony - tagging of days for -c and -s
 * 10/29/93 tony - -W flag for specifying line width
 * 07/29/98 tony - fix a malloc bug which 
 * 09/05/99 tony - redo some structures, fix leap year bug
 * 09/05/99 tony - add postscript output
 * 09/06/99 tony - add HTML output
 * 09/06/99 tony - add options to suppress holiday & birthday lists
 * 09/06/99 tony - add usage
 *
 * 03/02/00 version bigcal 2.1.0
 * 03/02/00 tony - fix two bugs evident only on first week of the month.
 *			title printing in postscript
 *			core dump with -w
 * 03/02/00 tony - if required, emit extra lines to show all events, even
 *			though that messes up the grid spacing.
 * 09/14/00 tony - fix bug with postscript month breaks when not showing weekends
 * 09/14/00 tony - add holiday shading
 * 10/18/00 tony - add repeating items
 * 11/09/2002 tony - add +b to tagging to get birthdays
 */

#ifndef lint
static	char *sccsid = "@(#)(aiuto) bigcal.c delta: %M% %I% %G% %U% build: %H% %T%";
#endif

#include <stdio.h>
#include <time.h>
#include <string.h>
#include <malloc.h>
#include <ctype.h>
/* #include <aiuto/libaiuto.h> */

/* System functions */
extern	char	*getenv();
extern	int	optind, getopt();
extern	char	*optarg;
extern	char	*strdup(const char *);

/* options */
static	int	bflag = 0;	/* print just blank form */
static	int	printWeekends = 0;	/* print weekends too */
static	int	weeks;		/* How many weeks to print */
static	int	edit = 0;	/* Edit calander option */
static	int	squeezeLines = 0;	/* squeeze blank lines */
static	int	postscript = 0;	/* print in postscript */
static	int	html = 0;	/* print as HTML */
static	int	noLoad = 0;	/* do not load .cal.birthdays & such */

static	int	ypos = 0;	/* y position on the page */
static	int	rowBaselineY;	/* y position at start of row */
static	int	cWidth = 0;

static	mdays[] = { 0,		/* really 1 based addressing */
	31, 28, 31, 30, 31, 30,
	31, 31, 30, 31, 30, 31
};

#define DAYS_PER_YEAR	365

/* Each file loaded adds a new Text instance to the chain of elements
 * for a day
 */
typedef struct _Text {
	struct _Text	*next;		/* single linked list */
	char		*text;		/* text to display */
	int		type;		/* what kind of thing */
} Text;

typedef struct {
	int	month;
	int	day;
	char	*tag;			/* text that goes next to the date */
	Text	*texts;			/* one Text for each line read, so
					 * you have one for .cal.birthdays
					 * and an additional one for .cal
					 */
	Text	*cur;			/* which item we are printing */
	char	holiday;		/* special holiday - print shaded */
} DayInfo;

static DayInfo	dayinfo[DAYS_PER_YEAR+32];

typedef struct {
	int	dow;		/* day of week */
	int	start;		/* begin time in HHMM - 24 hour */
	int	end;		/* end time */
	char	*what;		/* description */
	int	printed;	/* mark & sweep flag */
	int	prow;		/* internal to printRepeaters */
} Repeater;


#if !defined(MAX_REPEATS)
/* more than this and you are seriously overloaded */
#define MAX_REPEATS	30
#endif

static Repeater	repeaters[MAX_REPEATS];
static int	nrepeaters = 0;


int	dlevel;			/* debug level */
static	char	*home;		/* Users home directory */

#define	XMARGIN		36	/* 1/2 inch */
#define	YMARGIN		36	/* 1/2 inch */
#define	YTOP		756
#define	TEXTSIZE	7


#define	NORMAL		0
#define	HOLIDAY		1
#define	BIRTHDAY	2

static void getItems();
static void loadRepeaters(const char *);
static void printRepeaters();
static void printcal();
static void printhtml();
static void FindHome();
static void printTodoList();


static void usage()
{
	printf("bigcal %s\n", VERSION);
	printf("usage: bigcal [-beHnpstw] [-123456789qhy] [-W nn] [calanderFile ...]\n");
	printf("-b\tjust print a Blank calendar\n");
	printf("-e\tEdit $HOME/.cal file before printing calendar.\n");
	printf("-H\tprint calendar in Html.\n");
	printf("-n\tdo Not load $HOME/.cal, $HOME/.cal.birthdays & $HOME/.cal.base files.\n");
	printf("-p\tprint calendar in Postscript.\n");
	printf("-s\tSqueeze out blank lines.\n");
	printf("-t\tprint Todo list ($HOME/.todo).\n");
	printf("-w\tprint Weekend days.\n");
	printf("-W nn\tprint for terminal Width of nn.\n");

	printf("-q\tprint next Quarter of a year (13 weeks).\n");
	printf("-h\tprint next Half of a year (26 weeks).\n");
	printf("-y\tprint next Year (52 weeks).\n");
	printf("-123456789\tprint N weeks (default=4).\n");
	printf("\n");
	printf("Format of $HOME/.cal file:\n");
	printf("\t# comment\n");
	printf("\tmonth day [-tag] item1|item2 ...\n");
	printf("\t-tag items print on same line as the date\n");
	printf("\tother items print on successive lines.\n");
}

main(argc, argv)
	int	argc;
	char	*argv[];
{
	int	c;
	struct	tm	*tim;
	long	clock;
	int 	year, month, day;
	int	width;
	char	calFile[512];	/* Path to calander file */
	int	printTodo = -1;
	int	changedWeeks = 0;

	bflag = 0;
	printWeekends = 0;
	weeks = 4;
	width = 80;
	while((c = getopt(argc, argv, "bdehHnpqstwW:y123456789")) != EOF) {
		switch(c) {
		case 'b':	bflag = 1;	break;
		case 'd':	dlevel++;	break;
		case 'e':	edit = 1;	break;
		case 'H':
			html = 1;
			if(changedWeeks == 0) weeks = 26;
			if(printTodo < 0) printTodo = 0;
			break;
		case 'n':	noLoad = 1;	break;
		case 'p':
			postscript = 1;
			if(changedWeeks == 0) weeks = 16;
			if(printTodo < 0) printTodo = 0;
			break;
		case 's':
			squeezeLines = 1;	/* squeeze blank lines */
			break;
		case 't':
			printTodo = 1;	/* squeeze blank lines */
			break;
		case 'W':
			if(strcmp(optarg, "W") == 0) {
				width = 132;
			} else {
				width = atoi(optarg);
				if(width <= 20) width = 132;
			}
			break;
		case 'w':	printWeekends = 1;	break;

		case '1': case '2': case '3': case '4': case '5':
		case '6': case '7': case '8': case '9':
			weeks = c - '0';
			changedWeeks = 1;
			break;
		case 'q':				/* quarter */
			weeks = 13;
			changedWeeks = 1;
			if(printTodo < 0) printTodo = 0;
			break;
		case 'h':				/* half */
			weeks = 26;
			changedWeeks = 1;
			if(printTodo < 0) printTodo = 0;
			break;
		case 'y':				/* year */
			weeks = 52;
			changedWeeks = 1;
			if(printTodo < 0) printTodo = 0;
			break;

		default:
			fprintf(stderr, "%s: unknown switch %c\n", argv[0], c);
			usage();
			exit(1);
		}
	}

	FindHome();

	/* set up default month to do */
	time(&clock);
	tim = localtime(&clock);
	year = tim->tm_year + 1900;
	month = tim->tm_mon + 1;
	day = tim->tm_mday;

	/* adjust for leap years */
	if(month >= 3) {
		int y = year + 1;
		if(y % 4 == 0 && (y % 200 != 0) || (y % 1000 == 0)) {
			mdays[2] = 29;
		}
	} else {
		if(year % 4 == 0 && (year % 200 != 0) || (year % 1000 == 0)) {
			mdays[2] = 29;
		}
	}

	if(edit) {
		char	cmd[500];
		char	*editor = getenv("EDITOR");
		if(editor == NULL || *editor == '\0') editor = "vi";
		sprintf(cmd, "%s '+/^%d' %s/.cal", editor, month, home);
		system(cmd);
	}

	if(postscript) {
		if(printWeekends) {
			cWidth = 150 / 7;
		} else {
			cWidth = 150 / 5 ; 
		}
	} else {
		if(printWeekends) {
			cWidth = (width - 1) / 7 - 1;
		} else {
			cWidth = (width - 1) / 5 - 1; 
		}
	}

	/* Initialize and load calandars */
	for(c = 0; c < sizeof(dayinfo)/sizeof(dayinfo[0]); c++) {
		dayinfo[c].tag = NULL;
		dayinfo[c].texts = NULL;
		dayinfo[c].holiday = 0;
	}

	/* Last ones in print on top */
	if(noLoad == 0) {
		sprintf(calFile, "%s/.cal", home);
		getItems(calFile, month, day, NORMAL);
		sprintf(calFile, "%s/.cal.birthdays", home);
		getItems(calFile, month, day, BIRTHDAY);
		sprintf(calFile, "%s/.cal.holidays", home);
		getItems(calFile, month, day, HOLIDAY);
	}
	for( ;optind < argc; optind++) {
		getItems(argv[optind], month, day, NORMAL);
	}
	sprintf(calFile, "%s/.cal.regular", home);
	loadRepeaters(calFile);

	if(html) {
		printhtml(year, month, day, clock);
		exit(0);
	}

	printcal(year, month, day, clock);
	if(printTodo) printTodoList();

	exit(0);
}

static void printTodoList()
{
	char	todoFile[500];
	FILE	*fp;
	char	line[200];

	sprintf(todoFile, "%s/.todo", home);
	fp = fopen(todoFile, "r");
	if(fp == NULL) return;

	printf("\n");
	while(fgets(line, sizeof(line), fp)) {
		printf("%s", line);
	}
	fclose(fp);
}


int xpos(int col)
{
	if(printWeekends) {
		return XMARGIN + (72 * 7.5 * col / 7);
	} else {
		return XMARGIN + (72 * 7.5 * col / 5);
	}
}

int colwidth()
{
	if(printWeekends) {
		return (72 * 7.5 / 7);
	} else {
		return (72 * 7.5 / 5);
	}
}


static void emitPSHeader()
{
	printf("%!PS\n");
	printf("/fn /Helvetica findfont %d scalefont def\n", TEXTSIZE);
	printf("/fb /Helvetica-Bold findfont %d scalefont def\n",
		TEXTSIZE);
	printf("/fi /Helvetica-Italic findfont %d scalefont def\n",
		TEXTSIZE);
	printf("/shadowbox { newpath moveto 0 -%d rlineto %d 0 rlineto 0 %d rlineto closepath .8 setgray fill 0 setgray } def\n", 
		(int) (TEXTSIZE * 6.5),
		colwidth(),
		(int) (TEXTSIZE * 6.5));
}

void emitPSstring(const char *s)
{
	char	c;
	putc('(', stdout);
	while(c = *s++) {
		if((c == '(') ||
		   (c == ')') ||
		   (c == '\\')) {
			putc('\\', stdout);
		}
		putc(c, stdout);
	}
	putc(')', stdout);
}

/* print a shaded box
 * int x, y		lower left corner
 * int width, height	dimensions
 */

void PSshaded(int x, int y, int width, int height)
{
	printf(" %d %d newpath moveto", x, y);
	printf(" 0 %d rlineto", height);
	printf(" %d 0 rlineto", width);
	printf(" 0 -%d rlineto", height);
	printf(" closepath .8 setgray fill 0 setgray\n");
}


static void printcal(year, month, day, clock)
	int	year, month, day;
	long	clock;
{
	int	baseDate, dow, showMonth;
	int	cday;			/* index of day of month */
	int	week;			/* week # from base of calander */
	struct	tm	*tim;
	int	wStart, wEnd;	/* indices of which days to print */
	char	dbuf[50];
	int	monthBreak;

	showMonth = 0;
	wStart = 1;		/* Monday */
	wEnd = 5;		/* Friday */
	if(printWeekends) {
		wStart = 0;	/* Sunday */
		wEnd = 6;	/* Saturday */
	}
#define	NDAYS	(wEnd - wStart + 1)

	/* set up default month to do */
	time(&clock);
	tim = localtime(&clock);
	dow = tim->tm_wday;

	day -= dow;
	cday = day;
	/* we print header inside the loop for postscript */
	if(postscript) {
		emitPSHeader();
	} else {
		printf("Calendar for week of %d/%d/%d\n", month, day, year);
		printRepeaters();
	}

	for(week = 0; week < weeks; week++) {
		if(postscript && ((week % 16) == 0)) {
			ypos = YTOP;
			printf("%d %d moveto", XMARGIN, ypos);
			printf(" fb setfont (Calendar for week of %d/%d/%d) show\n",
						month, cday, year);
			ypos -= TEXTSIZE;
			printRepeaters();
			week += 1;
		}

		/* Print the divider between rows
		 *
		 */
		monthBreak = 8 - (cday + 6 - mdays[month]);
		if(monthBreak > 0 && monthBreak <= 8) {	/* month wrap */
			pline(wStart, wEnd, monthBreak);
		} else {
			pline(wStart, wEnd, 0);
		}

		ypos -= TEXTSIZE;
		/* Print the dates */
		baseDate = day + wStart;
		for(dow = 0; dow <= 6; dow++, day++, cday++) {
			if(cday < 1) {			/* before start */
				dbuf[0] = '\0';
			} else {
				if(cday > mdays[month]) { /* next month */
					cday = 1;
					month = (month % 12 ) + 1;
					showMonth = 1;
				}
				if(showMonth) {
					sprintf(dbuf, "%d/%d", month, cday);
				} else {			/* normal */
					sprintf(dbuf, "%d", cday);
				}
			}
			if(dow >= wStart && dow <= wEnd) {
				if(day >= 0 && dayinfo[day].tag) {
					strcat(dbuf, " ");
					strcat(dbuf, dayinfo[day].tag);
				}
				if(postscript) {
				    if(dayinfo[day].holiday) {
				    	printf(" %d %d shadowbox\n",
						xpos(dow-wStart), rowBaselineY);
				    }
				    printf("%d %d moveto",
						xpos(dow-wStart) + TEXTSIZE/2,
									ypos);
				    printf(" fb setfont (%s) show\n", dbuf);
				} else {
				    printf("| %-*.*s", cWidth-1, cWidth-1, dbuf);
				}
			}
		}
		if(postscript == 0) {
			printf("|\n");
		}
		ptext(baseDate, NDAYS);

		if(postscript && (((week % 16) == 15) || (week == weeks - 1))) {
			int	d;
			pline(wStart, wEnd, 0);
			/* Draw the vertical rules */
			for(d = 0; d <= NDAYS; d++) {
				printf("newpath %d %d moveto %d %d lineto stroke\n",
					xpos(d), ypos + TEXTSIZE/2,
					xpos(d), YTOP-TEXTSIZE);
			}
			printf("showpage\n");
		}
	}

	if(postscript == 0) {
		pline(wStart, wEnd, 0);
	}
}

static void FindHome()
{
	home = getenv("HOME");
	if(home == NULL) home = ".";
}



/* Get items for this month and the following month */

static void getItems(const char *file, int month, int day, int defaultItemType)
{
	char	line[1000];
	FILE	*cf;
	char	*m, *d, *t, *n, *tp;
	int	dmonth, dday, dayOfMonth;
	int	nextMonth, i;
	int	lineNumber = 0;

	if(bflag) return;

	cf = fopen(file, "r");
	if(cf == NULL) {
		return;
	}

	nextMonth = month % 12 + 1;

	while(fgets(line, sizeof(line), cf)) {
		DayInfo	*di;
		Text	*ti;
		int itemType = defaultItemType;

		lineNumber += 1;
		trimall(line);
		if(line[0] == '#' || line[0] == '\0') continue;

		m = strtok(line, " \t");
		dmonth = atoi(m);

		d = strtok(NULL, " \t");
		if(d == NULL) {
			fprintf(stderr, "%s: line %d:  - malformed date\n",
					file, lineNumber);
			continue;
		}
		if(strcmp(d, "*") == 0) {
			dayOfMonth = day;
		} else {
			dayOfMonth = atoi(d);
		}
		if(dayOfMonth < 1 || dayOfMonth > 31) {
			fprintf(stderr, "%s: Bad Date: %s %s\n", file, m, d);
			continue;
		}
		dday = dayOfMonth;
		t = strtok(NULL, "");
		if(t == NULL) t = "";

		/* shift future months down a little */
		if(dmonth != month) {
			for(i = month; i != dmonth; i = i % 12 + 1) {
				dday += mdays[i];
			}
		}
		if(dday >= sizeof(dayinfo)/sizeof(dayinfo[0])) {
			fprintf(stderr, "Too many days this year.\n");
			continue;
		}
		di = &dayinfo[dday];
		di->month = dmonth;
		di->day = dayOfMonth;

		if(*t == '+' && (*(t+1) == 'h' || *(t+1) == 'H')) {
			di->holiday = 1;
			t += 2;
			while(*t && isspace(*t)) t++;
		}
		if(*t == '+' && (*(t+1) == 'b' || *(t+1) == 'B')) {
			itemType = BIRTHDAY;
			t += 2;
			while(*t && isspace(*t)) t++;
		}
		if(*t == '-') {
			char	tag[50];
			t++;
			tp = tag;
			while(*t && !isspace(*t)) {
				char c = *t++;
				if(c == '_') *tp++ = ' ';
				else *tp++ = c;
			}
			*tp = '\0';
			if(di->tag) {
				n = malloc(strlen(di->tag)+strlen(tag)+2);
				strcpy(n, di->tag);
				strcat(n, ",");
				strcat(n, tag);
				free(di->tag);
				di->tag = n;
			} else {
				di->tag = strdup(tag);
			}
			/* printf("Tag %d = '%s'\n", dday, di->tag); */
		}

		while(isspace(*t) || *t == '|') t++;

		ti = (Text *) malloc(sizeof(Text));
		ti->text = strdup(t);
		ti->next = di->texts;
		di->texts = ti;
		di->cur = ti;
		ti->type = itemType;
	}
	fclose(cf);
}



/* Draw the divider line between rows
 *	int	start;		index to start on (0 or 1)
 *	int	end;		index to end (5 or 6)
 *	int	split;		index of last day of month
 */

pline(int start, int end, int split)
{
	int	ndots, day;
	static	int	lastSplit = 0;
	static	int	lastypos = 0;
	char	sep;

	rowBaselineY = ypos;

	if(postscript) {
		int	nDays = end - start + 1;	/* how many days */
		printf("newpath %d %d moveto", xpos(0), ypos);

		if(lastSplit > 0) {
			/* draw the bold line on the left */
			printf(" %d %d lineto \n",
					xpos(lastSplit), ypos);
			if(lastSplit < end) {
				/* draw a bold ascender up at the split */
				printf(" %d %d lineto stroke",
						xpos(lastSplit),lastypos);
			}
			printf(" 3 setlinewidth stroke\n");

			/* draw the thin line on the right */
			printf("newpath %d %d moveto", xpos(lastSplit), ypos);
			printf(" %d %d lineto 1 setlinewidth stroke\n",
					xpos(nDays), ypos);

			lastSplit = 0;
		} else {
			if(split == 0) {
				printf(" %d %d lineto 1 setlinewidth stroke\n",
						xpos(nDays), ypos);
			} else {
				split = split - start - 1;
				if(split > nDays) split = nDays;
				printf(" %d %d lineto 1 setlinewidth stroke\n",
						xpos(split), ypos);
				printf("newpath %d %d moveto", xpos(split),
								ypos);
				printf(" %d %d lineto 3 setlinewidth stroke\n",
						xpos(nDays), ypos);
				lastSplit = split;
			}
		}
		lastypos = ypos;
		ypos -= TEXTSIZE / 2;
		return;
	}

	start += 1;		/* make 1 based */
	end += 1;
	if(lastSplit) {
		for(day = start; day <= end; day++) {
			fputc('+', stdout);
			if(day < lastSplit) {
				sep = '=';
			} else {
				sep = '-';
			}
			ndots = cWidth;
		while(ndots-- > 0) fputc(sep, stdout);
		}
		printf("+\n", sep);
		lastSplit = 0;
		return;
	}

	for(day = start; day <= end; day++) {
		fputc('+', stdout);
		if(split == 0 || day < split) {
			sep = '-';
		} else {
			sep = '=';
		}
		ndots = cWidth;
		while(ndots-- > 0) fputc(sep, stdout);
	}
	printf("+\n", sep);

	if(split > 0) {
		lastSplit = split;
	}
}


/* print the contents of the cell */
ptext(int baseDate, int nDays)
{
	int	line, dow, k, day;
	char	tbuf[100], *tb, c;
	int	nrows = 4;
	int	curfont = NORMAL;
	int	font;
	int	moreItems = 0;

	if(postscript) {
		nrows = 6;
		printf("fn setfont\n");
	}

	for(line = 1; line < nrows || (moreItems != 0); line++) {
		DayInfo	*di;
		int	haveText = 0;

		ypos -= TEXTSIZE;
		moreItems = 0;
		for(dow = 0; dow < nDays; dow++) {
			day = baseDate + dow;
			di = &dayinfo[day];

			if(day < 1 ||
			   day >= sizeof(dayinfo)/sizeof(dayinfo[0]) ||
			   di->cur == NULL) {
				if(postscript == 0) printf("|%*s", cWidth, "");
				continue;
			}

				/* Copy next item to line buffer */

			/* Trim leading delimiters */
			while(isspace(*di->cur->text) ||
			      *di->cur->text == '|') {
				di->cur->text++;
				if(*di->cur->text == '\0') {
					/* at end of string traverse to next
					 * text block
					 */
					di->cur = di->cur->next;
				}
			}

			/* if no more text, just print blanks */
			if(di->cur == NULL) {
				if(postscript == 0) printf("|%*s", cWidth, "");
				continue;
			}

			font = di->cur->type;
			/* copy the characters out */
			tb = tbuf;
			for(k = 0; k < cWidth; k++) {
				if(*di->cur->text) {
					haveText = 1;
					*tb = *di->cur->text++;
					if(*tb == '|') break;
					tb++;
				} else {
					/* go to next Text at end of string */
					di->cur = di->cur->next;
					break;
				}
			}
			*tb = '\0';
			if(postscript) {
				if(*tbuf) {
					if(font != curfont) {
						if(font == NORMAL) {
							printf("fn setfont\n");
						} else if(font == BIRTHDAY) {
							printf("fi setfont\n");
						} else if(font == HOLIDAY) {
							printf("fb setfont\n");
						}
						curfont = font;
					}
			    		printf("%d %d moveto ",
						xpos(dow) + TEXTSIZE/2, ypos);
					emitPSstring(tbuf);
			    		printf(" show\n");
				}
			} else {
				printf("|%-*.*s", cWidth, cWidth, tbuf);
			}

			/* Check for more items left to print */
			if(di->cur != NULL && *di->cur->text != '\0') {
				moreItems = 1;
			}
		}
		if(postscript == 0) {
			printf("|\n");
		}
		if(haveText == 0 && squeezeLines > 0) break;
	}
}

/* Do a calander in HTML.   Someday, this should be unified into
 * a single model for output
 */
static void printhtml(year, month, day, clock)
	int	year, month, day;
	long	clock;
{
	int	dow;
	int	week;
	struct	tm	*tim;
	int	wStart, wEnd;	/* indices of which days to print */
	int	haveText;
	char	*colors[] = { "#ffffff", "#cccccc"};
#define	NCOLORS (sizeof(colors)/sizeof(colors[0]))
	int	bgcolor = 0;

	wStart = 1;		/* Monday */
	wEnd = 5;		/* Friday */
	if(printWeekends) {
		wStart = 0;	/* Sunday */
		wEnd = 6;	/* Saturday */
	}
#define	NDAYS	(wEnd - wStart + 1)

	/* set up default month to do */
	time(&clock);
	tim = localtime(&clock);
	dow = tim->tm_wday;

	printf("<HTML>\n");
	printf("<HEAD>\n");
	printf("<TITLE>Calendar for %d/%d/%d</TITLE>\n", month, day, year);
	printf("</HEAD>\n");
	printf("<BODY>\n");

	printf("<TABLE BORDER=1>\n");

	day -= dow;
	for(week = 0; week < weeks; week++) {

		printf("<TR>");

		/* Print the dates */
		for(dow = 0; dow <= 6; dow++, day++) {
			Text	*ti;
			DayInfo	*di = &dayinfo[day];
			char	dbuf[50];
			int	nRows = 0;

			if(dow < wStart || dow > wEnd) continue;

			if(di->month != month) {
				month = di->month;
				bgcolor = (bgcolor + 1) % NCOLORS;
			}
			sprintf(dbuf, "%d/%d", di->month, di->day);
			if(di->tag) {
				strcat(dbuf, " ");
				strcat(dbuf, di->tag);
			}
			printf("<TD VALIGN=TOP BGCOLOR=\"%s\"><B>%s</B>",
						colors[bgcolor], dbuf);

			haveText = 0;
			for(ti = di->texts; ti; ti = ti->next) {
				char *etag = "";
				char *tp;
				haveText += 1;
				printf("<BR>");
				if(ti->type == BIRTHDAY) {
					printf("<I>");
					etag = "</I>";
				} else if(ti->type == HOLIDAY) {
					printf("<B>");
					etag = "</B>";
				}
				for(tp = ti->text; *tp; tp++) {
					if(*tp == '|') {
						printf("<BR>");
					} else {
						fputc(*tp, stdout);
					}
				}
				fputs(etag, stdout);
			}
			if(nRows < haveText) nRows = haveText;
			while(nRows < 5 && dow == wEnd) {
				printf("<BR>");
				nRows += 1;
			}
			printf("</TD>\n");
		}
		printf("</TR>\n");
	}
	printf("</TABLE>\n");
	printf("</BODY>\n");
	printf("</HTML>\n");

}

static void loadRepeaters(const char *fname)
{
	char	line[300];

	FILE *inp = fopen(fname, "r");
	if(inp == NULL) {
		return;
	}

	while(fgets(line, sizeof(line), inp) != NULL) {
		char	*tok;
		trimall(line);
		if(line[0] == '#' || line[0] == '\0') continue;

		if(nrepeaters > sizeof(repeaters) / sizeof(repeaters[0])) {
			fprintf(stderr, "too many repeating events.  Max is %d\n",
				sizeof(repeaters) / sizeof(repeaters[0]));
			break;
		}
		tok = strtok(line, " \t");
		repeaters[nrepeaters].dow = atoi(tok);
		tok = strtok(NULL, " \t");
		repeaters[nrepeaters].start = atoi(tok);
		tok = strtok(NULL, " \t");
		repeaters[nrepeaters].end = atoi(tok);
		tok = strtok(NULL, "");
		if(tok != NULL) {
			repeaters[nrepeaters].what = strdup(tok);
			nrepeaters += 1;
		}
	}
	fclose(inp);
}

/* Print all the regularly repeating items */
static void printRepeaters()
{
	int	line, ir, dow;
	int	nrows = 4;
	int	wstart, wend;

	if(printWeekends) {
		wstart = 0;
		wend = 6;
	} else {
		wstart = 1;
		wend = 5;
	}
	pline(wstart, wend, 0);

	if(postscript) {
		nrows = 6;
		printf("fn setfont\n");
	}
	for(ir = 0; ir < nrepeaters; ir++) {
		Repeater *r = &repeaters[ir];
		r->printed = 0;
		/* compute row the item should be on */
		if(r->start < 800) {
			r->prow = 0;
		} else {
			r->prow = (r->start - 800) * nrows / 1200;
		}
		if(r->prow >= nrows) r->prow = nrows - 1;
	}
	for(ir = 0; ir < nrepeaters; ir++) {
		int irx, nafter = 0;
		Repeater *r = &repeaters[ir];
		for(irx = ir+1; irx < nrepeaters; irx++) {
			if(repeaters[irx].dow != r->dow) continue;
			nafter += 1;
		}
		if(r->prow + nafter >= nrows) {
			r->prow = nrows - nafter - 1;
		}
	}

	for(line = 1; line < nrows; line++) {

		ypos -= TEXTSIZE;
		for(dow = wstart; dow <= wend; dow++) {
			for(ir = 0; ir < nrepeaters; ir++) {
				char	tbuf[300], *rwhat;
				int	prow;
				Repeater *r = &repeaters[ir];

				if(r->dow != dow) continue;
				if(r->printed) continue;
				/* compute row the item should be on */
				if(r->prow > line) continue;

				r->printed = 1;
				sprintf(tbuf, "%4d-%4d %-100.100s",
					r->start, r->end, r->what);

				if(postscript) {
					int xbase = xpos(dow-wstart);
					int ybase = ypos + TEXTSIZE/2;
					int shadedHeight = TEXTSIZE;
					if(line == nrows-1) {
						ybase -= TEXTSIZE / 2;
						shadedHeight += TEXTSIZE/2;
					}
					PSshaded(xbase, ybase,
						colwidth(), shadedHeight);
			    		printf("%d %d moveto ",
						xbase + TEXTSIZE/2,
						ypos + TEXTSIZE/2);
					emitPSstring(tbuf);
			    		printf(" show\n");
				} else {
					printf("|%-*.*s", cWidth, cWidth, tbuf);
				}
				goto nextdow;
			}
nextdow:
		}
		if(postscript == 0) {
			printf("|\n");
		}
	}
}
